import socket
import sys
import struct
import threading

#
# this example shows how to register for events of the CODESYS control runtime
# the mechanism uses unix domain sockets
# 
#
# client                                        server 
#   |                                      (codesys runtime)
#   |    register req id, cmp id, event id ->     |
#   |    <- ACK                                   |
#   ~                                             |

#   ~
#   |    <- TRIGGER (req id)                      |
#   ~

#   ~
#   |    <- TRIGGER (req id)                      |
#   ~
#
# - any invalid message or protocol violation will lead to a closed socket
# 


# This class helps to connect to a CODESYS Control runtime
# and listen to the internal event mechanism of CODESYS runtimes
class CODESYS_EventHandler(threading.Thread):

    # 
    # msg register: 
    # client register at server: you have to send this to the server to register for a certain event
    # - unsigned int (4 bytes) :    request ID -> use sequential or unique number
    # - int (4 bytes) :             message type: MSGTYPE_REGISTER
    # - unsigned int (4 bytes) :    component ID
    # - unsigned int (4 bytes) :    event ID
    #
    # hints:
    #  1) registration is only done once per client and only at startup
    #  2) one client can only register for one event
    #  3) you can only try to register once (if it fails, the socket is closed, try again)
    #
    # response
    # you will get a response for your registration:
    # - unsigned int (4 bytes) :    response ID (from given request ID)
    # - int (4 bytes) :             message type: MSGTYPE_ACK
    # - unsigned int (4 bytes) :    error code (see above)


    # msg trigger:
    # the server sends a trigger message each time the registered event occurs. Event must be registered before
    # - unsigned int (4 bytes) :    response ID (from given request ID)
    # - int (4 bytes) :             message type: MSGTYPE_TRIGGER
    # respone: None
    
    # msg trigger_state:
    #  the server sends a trigger message each time the registered event occurs. Event must be registered before
    #  Additionally the message contains the application name and application state
    # - unsigned int (4 bytes) :    response ID (from given request ID)
    # - int (4 bytes) :             message type: MSGTYPE_TRIGGER_STATE
    # - char (61 bytes) :           name of application
    # - uint (4 bytes) :            state of application
    # respone: None

    # server endpoint for events:
    endpoint = '/var/run/codesysextension/events.sock'

    # message types
    MSGTYPE_ACK=0
    MSGTYPE_REGISTER=1    
    MSGTYPE_TRIGGER=2
    MSGTYPE_TRIGGER_STATE=3

    # constants
    APPLICATION_NAME_MAX_LENGTH=61

    # error codes
    ERR_OK=0x0000             # ok, registration succeeded
    ERR_NO_MEMORY=0x0011      # registration failed, not enough memory
    ERR_NO_OBJECT=0x0010      # registration failed, component and event ID is not a valid combination

    def __init__(self, componentid, eventid, callbackfunction):
        self.sock = None
        self.callbackfunction = callbackfunction
        self.componentid = componentid
        self.eventid = eventid        

        threading.Thread.__init__(self)

    def run(self):
        error = ""

        if not callable(self.callbackfunction):
            print("Callback function is not callable")
            return
        
        try:            
            # create unix domain socket
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            # and connect to CODESYS runtime (event server)
            print("connecting to runtime event server: " + CODESYS_EventHandler.endpoint)
            self.sock.connect(CODESYS_EventHandler.endpoint)
        except socket.error as msg:
            print(msg)
            return False

        try:
            # register for given event
            requestId = 0
            print("register for event ({},{})".format(self.componentid, self.eventid))
            format = "IiII"
            request = struct.pack(format, requestId, CODESYS_EventHandler.MSGTYPE_REGISTER, self.componentid, self.eventid)
            self.sock.sendall(request)

            # and check response
            #print("waiting for response")
            response = self.sock.recv(12)
            format = "IiI"
            responseId, msgType, result  = struct.unpack(format, response)
            #print("response: " + str((responseId, msgType, result)))
            if result == CODESYS_EventHandler.ERR_OK:
                print("registration succeeded")                
            elif result == CODESYS_EventHandler.ERR_NO_OBJECT:
                print("error: no such event!")
                return
            elif result == CODESYS_EventHandler.ERR_NO_MEMORY:
                print("error: no more connections possible")
                return
            print("waiting for event")
            while True:
                trigger = self.sock.recv(76)
                if not trigger:
                    print("error: recv failed, done")
                    break
                else:
                    format = "Ii"
                    responseId, msgType  = struct.unpack(format, trigger[:8])
                    if responseId == requestId:
                        if msgType == self.MSGTYPE_TRIGGER_STATE:
                            format = "Ii" + str(self.APPLICATION_NAME_MAX_LENGTH) + "sI"
                            responseId, msgType, applicationName, applicationState  = struct.unpack(format, trigger)
                            applicationName = bytearray(applicationName).decode()
                            self.callbackfunction(self.componentid, self.eventid, applicationName, applicationState)
                        else:
                            self.callbackfunction(self.componentid, self.eventid)
                
            
        except socket.error as msg:
            print("error: " + str(msg))
        finally:
            if self.sock != None:
                self.sock.close()
# end of class


# list of usefull events
liEvents = {
    "CmpApp_EVT_StartDone" : (0x00000002, 0x10000 + 2),
    "CmpApp_EVT_StopDone" : (0x00000002, 0x10000 + 4),
    "CmpApp_EVT_ResetDone" : (0x00000002, 0x10000 + 6),    
    "CmpApp_EVT_ExitDone" : (0x00000002, 0x10000 +  15),
    "CmpApp_EVT_AllBootprojectsLoaded" : (0x00000002, 0x10000 +  25),    
    "CmpApp_EVT_StateChanged" : (0x00000002, 0x10000 + 43),
    "CmpApp_EVT_CmpApp_Exception" : (0x00000002, 0x00080000 + 28),
    "CmpDevice_EVT_CmpDevice_InteractiveLogin" : (0x0000000E, 0x00010000 + 1),
}

# list of states
liStates = {
    0 : "NONE",
    1 : "RUN",
    2 : "STOP", 
    3 : "DEBUG_HALT_ON_BP",
    4 : "AS_DEBUG_STEP",
    5 : "SINGLE_CYCLE",
}


# functions that should be called if an event occurs
def callbackfunction_start(componentid, eventid):
    print("Start event occured")

def callbackfunction_stop(componentid, eventid):
    print("Stop event occured")

def callbackfunction_exception(componentid, eventid):
    print("Exception event occured")

def callbackfunction_wink(componentid, eventid):
    print("Wink event occured")

def callbackfunction_statechanged(componentid, eventid, applicationName, applicationState):
    print("application: "+ str(applicationName) + "  State: "+ str(liStates.get(applicationState, "UNDEFINED")))

# example for event handlers:
if __name__ == "__main__":

    componentid, eventid = liEvents["CmpApp_EVT_StartDone"]
    myEventHandler_Start = CODESYS_EventHandler(componentid, eventid, callbackfunction_start)
    myEventHandler_Start.start()

    componentid, eventid = liEvents["CmpApp_EVT_StopDone"]
    myEventHandler_Stop = CODESYS_EventHandler(componentid, eventid, callbackfunction_stop)
    myEventHandler_Stop.start()

    componentid, eventid = liEvents["CmpApp_EVT_CmpApp_Exception"]
    myEventHandler_Exception = CODESYS_EventHandler(componentid, eventid, callbackfunction_exception)
    myEventHandler_Exception.start()

    componentid, eventid = liEvents["CmpDevice_EVT_CmpDevice_InteractiveLogin"]
    myEventHandler_Exception = CODESYS_EventHandler(componentid, eventid, callbackfunction_wink)
    myEventHandler_Exception.start()

    componentid, eventid = liEvents["CmpApp_EVT_StateChanged"]
    myEventHandler_StateChange = CODESYS_EventHandler(componentid, eventid, callbackfunction_statechanged)
    myEventHandler_StateChange.start()
